{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "16e9891c-708b-41fc-8a2a-9f375f9ae58d",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Original challenge.py\n",
    "import hashlib\n",
    "import struct\n",
    "\n",
    "class CustomSHA:\n",
    "    def __init__(self, data: bytes):\n",
    "        self.data: bytes = data\n",
    "\n",
    "    def process(self) -> str:\n",
    "        h0 = 0x674523EFCD.to_bytes(5, byteorder='big')\n",
    "        h = h0\n",
    "        data = self.preprocess()\n",
    "\n",
    "        for i in range(0, len(data), 8):\n",
    "            chunk = data[i:i+8]\n",
    "\n",
    "            h = hashlib.sha256(chunk + h).digest()[:5]\n",
    "\n",
    "        return f'{h.hex()}'\n",
    "\n",
    "    def preprocess(self) -> bytes:\n",
    "        data = self.data\n",
    "        data = bytearray(data)\n",
    "\n",
    "        data.append(0x80)\n",
    "        while len(data) % 8 != 0:\n",
    "            data.append(0x00)\n",
    "\n",
    "        data += struct.pack('>Q', 8 * len(self.data))\n",
    "\n",
    "        return bytes(data)\n",
    "\n",
    "\n",
    "    def sha_custom(self):\n",
    "        return self.process()\n",
    "\n",
    "with open(\"Hack.zip\", \"rb\") as f:\n",
    "    data_hack = f.read()\n",
    "    CustomSHA(data_hack)\n",
    "    sha_custom_hack = CustomSHA(data_hack).sha_custom()\n",
    "\n",
    "def challenge(received_json):\n",
    "    response_json = {}\n",
    "    if 'action' in received_json:\n",
    "        if received_json['action'] == 'hash':\n",
    "            if 'data' in received_json:\n",
    "                data = bytes.fromhex(received_json['data'])\n",
    "                if data == data_hack:\n",
    "                    response_json['error'] = \"Forbidden data\"\n",
    "                else:\n",
    "                    sha_custom = CustomSHA(data)\n",
    "                    response_json['hash'] = sha_custom.sha_custom()\n",
    "\n",
    "                    if sha_custom_hack == response_json['hash']:\n",
    "                        response_json['flag'] = FLAG\n",
    "            else:\n",
    "                response_json['error'] = 'No data provided'\n",
    "        else:\n",
    "            response_json['error'] = 'Invalid action'\n",
    "    else:\n",
    "        response_json['error'] = 'No action provided'\n",
    "\n",
    "    return response_json\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "id": "f26d685a-024a-4991-93e0-23c15cb14f2a",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "d8e3e7c169\n"
     ]
    }
   ],
   "source": [
    "print(sha_custom_hack)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "id": "3d192049-dadf-4153-8bc1-0fa24d87eb8b",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Compute all the intermediate hashes h_i for a given data string\n",
    "def compute_chain_h(data):\n",
    "    data = bytearray(data)\n",
    "    hs = []\n",
    "\n",
    "    h0 = 0x674523EFCD.to_bytes(5, byteorder='big')\n",
    "    h = h0\n",
    "    hs.append(h0)\n",
    "\n",
    "    for i in range(0, len(data), 8):\n",
    "        chunk = data[i:i+8]\n",
    "\n",
    "        h = hashlib.sha256(chunk + h).digest()[:5]\n",
    "        hs.append(h)\n",
    "\n",
    "    return hs\n",
    "\n",
    "hshack = compute_chain_h(data_hack)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "id": "68a5ea79-d8c2-49bb-b4e5-a58c33374660",
   "metadata": {},
   "outputs": [],
   "source": [
    "\"\"\"\n",
    "Finds a buffer of size length, such that it hashes to one of the values in hs (starting from hstart)\n",
    "\"\"\"\n",
    "def find_collision(hstart, hs, length = 1):\n",
    "    buf = bytearray()\n",
    "    hset = set(hs)\n",
    "\n",
    "    h = hstart\n",
    "    i = 0\n",
    "    for i in range(length-1):\n",
    "        chunk = bytearray(8)\n",
    "        buf += chunk\n",
    "        i += 1\n",
    "        h = hashlib.sha256(chunk + h).digest()[:5]\n",
    "\n",
    "    i=0\n",
    "    while True:\n",
    "        chunk = bytearray(i.to_bytes(8, byteorder='big'))\n",
    "        i += 1\n",
    "        h2 = hashlib.sha256(chunk + h).digest()[:5]\n",
    "        if h2 in hset:\n",
    "            buf += chunk\n",
    "            return buf, hs.index(h2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "id": "ecc6a501-7be8-449d-9428-d09b1f80156f",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "-3367\n",
      "-2767\n",
      "3435\n",
      "-4605\n",
      "-7649\n",
      "-7345\n",
      "245\n",
      "-5974\n",
      "-8524\n",
      "-689\n",
      "7603\n",
      "4968\n",
      "-8475\n",
      "-8854\n",
      "2893\n",
      "-9943\n",
      "-4385\n",
      "-779\n"
     ]
    }
   ],
   "source": [
    "# Now, we find several independent modifications of the file that reduces/extends its size\n",
    "# We want them to apply to disjoint sets of (h_i) to be able to apply them independently later\n",
    "intervals = []\n",
    "index0 = 0\n",
    "for j in range(18):\n",
    "    buf, index = find_collision(hshack[index0], hshack[index0:index0+20000], length=10000)\n",
    "    intervals.append((index0, index0+index, buf))\n",
    "    index0 += index\n",
    "    print(25000-index)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "id": "309f646d-d4b6-45bd-88f9-66f02ee47fe0",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[1, 2, 10, 11, 13, 16]\n"
     ]
    }
   ],
   "source": [
    "# In this code, we look for a subset of modifications that does not alter the file size\n",
    "# To do so, we solve the Subset Sum problem for the above integers.\n",
    "diffs = []\n",
    "for (i0, i1, buf) in intervals:\n",
    "    diffs.append(len(buf)//8 - (i1-i0))\n",
    "\n",
    "sums = set()\n",
    "sols = dict()\n",
    "for i, v in enumerate(diffs):\n",
    "    for u in list(sums):\n",
    "        sums.add(u+v)\n",
    "        sols[u+v] = sols[u].copy()\n",
    "        sols[u+v].add(i)\n",
    "    sums.add(v)\n",
    "    sols[v] = set([i])\n",
    "    \n",
    "sol = list(sols[0])\n",
    "sol.sort()\n",
    "print(sol)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "id": "93be6425-7f56-4a49-a557-673d0d26070d",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "e2c3cc7202 e2c3cc7202\n",
      "1922930.125 1927315.125\n",
      "e2c3cc7202 e2c3cc7202\n",
      "1914076.125 1927315.125\n",
      "e2c3cc7202 e2c3cc7202\n",
      "1919044.125 1927315.125\n",
      "e2c3cc7202 e2c3cc7202\n",
      "1926647.125 1927315.125\n",
      "e2c3cc7202 e2c3cc7202\n",
      "1930082.125 1927315.125\n",
      "e2c3cc7202 e2c3cc7202\n",
      "1927315.125 1927315.125\n"
     ]
    }
   ],
   "source": [
    "# Now, we apply the selected modifications to Hack.zip\n",
    "def replace_chain(data, v):\n",
    "    (i0, i1, buf) = v\n",
    "    databis = data[:i0*8] + buf + data[i1*8:]\n",
    "    hsbis = compute_chain_h(databis)\n",
    "    print(hsbis[-1].hex(), hshack[-1].hex())\n",
    "    print(len(databis)/8, len(data_hack)/8)\n",
    "    return databis\n",
    "\n",
    "datafin = data_hack[:]\n",
    "for i in sol[::-1]:\n",
    "    datafin = replace_chain(datafin, intervals[i])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "id": "c640c511-ec74-4cb2-adae-23829a7f31b8",
   "metadata": {},
   "outputs": [],
   "source": [
    "# And we produce the string expected by challenge.py\n",
    "data_hack2 = bytes(datafin)\n",
    "\n",
    "import json\n",
    "\n",
    "j = {\n",
    "    \"action\":\"hash\",\n",
    "    \"data\": data_hack2.hex()\n",
    "}\n",
    "\n",
    "# Serializing json\n",
    "json_object = json.dumps(j)\n",
    "with open(\"response.json\", \"w\") as outfile:\n",
    "    outfile.write(json_object)"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "projectname",
   "language": "python",
   "name": "projectname"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.6"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
